/*
 * Copyright (C) 2000 drscholl@users.sourceforge.net
 * This is free software * distributed under the terms of the GNU Public
 * License.  See the file COPYING for details.
 * 
 * $Id: libpoll.c,v 1.26 2000/10/25 17:34:40 drscholl Exp $
 */

#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "libpoll.h"
#include "debug.h"

#if HAVE_KQUEUE
#define check_fd(p) (p)->fd
#elif HAVE_POLL
#define check_fd(p) lp->poll_fd[(p)->pollfd_offset].fd
#else
#define check_fd(p) (p)->fd
#endif

#if DEBUG
static int
lp_validate_ctx (lp_ctx * lp)
{
    ASSERT_RETURN_IF_FAIL (debug_valid (lp, sizeof (lp_ctx)), 0);
#if HAVE_POLL && !defined(HAVE_KQUEUE)
    ASSERT_RETURN_IF_FAIL (lp->num_fds <= lp->max_fds, 0);
    ASSERT_RETURN_IF_FAIL (lp->max_fds == 0 ||
			   debug_valid (lp->poll_fd,

					sizeof (struct pollfd) * lp->max_fds),
			   0);
#endif
    ASSERT_RETURN_IF_FAIL (debug_valid (lp->bucket,
					sizeof (lp_fd_ctx *) *
					lp->num_buckets), 0);
    return 1;
}

static int
lp_validate_fd_ctx (lp_fd_ctx * fd)
{
    ASSERT_RETURN_IF_FAIL (debug_valid (fd, sizeof (lp_fd_ctx)), 0);
    return 1;
}
#endif

static lp_fd_ctx *
find_fd (lp_ctx * lp, int fd)
{
    lp_fd_ctx *bucket;

    ASSERT (lp_validate_ctx (lp));
    bucket = lp->bucket[fd % lp->num_buckets];
    while (bucket)
    {
	ASSERT (lp_validate_fd_ctx (bucket));
#if HAVE_POLL && !defined(HAVE_KQUEUE)
	ASSERT (bucket->pollfd_offset < lp->num_fds);
	if (fd == lp->poll_fd[bucket->pollfd_offset].fd)
	    return bucket;
#else
	if (bucket->fd == fd)
	    return bucket;
#endif
	bucket = bucket->next;
    }
    return 0;
}

lp_ctx *
lp_init (void)
{
    lp_ctx *lp = CALLOC (1, sizeof (lp_ctx));

#if HAVE_KQUEUE
    lp->kq_fd = kqueue ();
    if (lp->kq_fd == -1)
    {
	perror ("lp_init: kqueue");
	FREE (lp);
	return 0;
    }
#endif
    /* initialize hash table */
    lp->num_buckets = 1027;
    lp->bucket = CALLOC (lp->num_buckets, sizeof (lp_fd_ctx *));

    return lp;
}

void
lp_zero (lp_ctx * lp)
{
    int     i;
    lp_fd_ctx *bucket;

    ASSERT (lp_validate_ctx (lp));
#if HAVE_KQUEUE
    lp->num_changes = 0;

    for (i = 0; i < lp->num_buckets; i++)
    {
	bucket = lp->bucket[i];
	while (bucket)
	{
	    ASSERT (lp_validate_fd_ctx (bucket));
	    if (bucket->flags & LP_READ)
	    {
		if (lp->num_changes == lp->max_changes)
		{
		    lp->max_changes += 10;
		    lp->kq_change = REALLOC (lp->kq_change,
					     sizeof (struct kevent) *

					     lp->max_changes);
		}
		lp->kq_change[lp->num_changes].ident = bucket->fd;
		lp->kq_change[lp->num_changes].filter = EVFILT_READ;
		lp->kq_change[lp->num_changes].flags = EV_DELETE;
		lp->num_changes++;
	    }
	    if (bucket->flags & LP_WRITE)
	    {
		if (lp->num_changes == lp->max_changes)
		{
		    lp->max_changes += 10;
		    lp->kq_change = REALLOC (lp->kq_change,
					     sizeof (struct kevent) *

					     lp->max_changes);
		}
		lp->kq_change[lp->num_changes].ident = bucket->fd;
		lp->kq_change[lp->num_changes].filter = EVFILT_WRITE;
		lp->kq_change[lp->num_changes].flags = EV_DELETE;
		lp->num_changes++;
	    }
	    bucket = bucket->next;
	}
    }

    /* flush queue */
    if (kevent (lp->kq_fd, lp->kq_change, lp->num_changes, 0, 0, 0) == -1)
	perror ("lp_zero: kevent");

    lp->num_changes = 0;
#endif

    /* free up the hash table */
    for (i = 0; i < lp->num_buckets; i++)
    {
	bucket = lp->bucket[i];
	while (bucket)
	{
	    lp_fd_ctx *tmp = bucket;

	    ASSERT (lp_validate_fd_ctx (bucket));
	    bucket = bucket->next;
	    FREE (tmp);
	}
	lp->bucket[i] = 0;
    }
#if !defined(HAVE_KQUEUE) && !defined(HAVE_POLL)
    FD_ZERO (&lp->read_fds);
    FD_ZERO (&lp->write_fds);
    lp->max_fd = 0;
#endif
    lp->num_fds = 0;
}

void
lp_set (lp_ctx * lp, int fd, int flags, void *data)
{
    lp_fd_ctx *bucket;

#if HAVE_POLL && !defined(HAVE_KQUEUE)
    struct pollfd *pfd;
#endif

    ASSERT (lp_validate_ctx (lp));

    /* look up the context associated with this file descriptor (if present) */
    bucket = find_fd (lp, fd);

    /* no context yet, create a new one */
    if (!bucket)
    {
	bucket = CALLOC (1, sizeof (lp_fd_ctx));
#if HAVE_POLL && !defined(HAVE_KQUEUE)
	/* allocate a new struct */
	if (lp->num_fds == lp->max_fds)
	{
	    lp->max_fds += 128;
	    lp->poll_fd =

		REALLOC (lp->poll_fd, sizeof (struct pollfd) * lp->max_fds);
	}
	/* zero the contents since it could be junk */
	lp->poll_fd[lp->num_fds].fd = fd;
	lp->poll_fd[lp->num_fds].events = 0;
	lp->poll_fd[lp->num_fds].revents = 0;
	/* save the offset into the poll_fd[] array so we can refer to it
	 * later.  note that we use an index into the array rather than a
	 * pointer because poll_fd[] gets realloc'd when the array becomes
	 * full.
	 */
	bucket->pollfd_offset = lp->num_fds;
#else
	/* all we need is to save the fd so we can handle collisions in
	 * the hash table.
	 */
	bucket->fd = fd;
#endif
	/* add to the hash table */
	bucket->next = lp->bucket[fd % lp->num_buckets];
	lp->bucket[fd % lp->num_buckets] = bucket;

	/* note how many fds we are watching */
	lp->num_fds++;
    }

    /* ensure the context we have is valid */
    ASSERT (lp_validate_fd_ctx (bucket));

    /* save the current state of this fd */
    bucket->flags |= (flags & LP_RDWR);

#if HAVE_KQUEUE
    if (flags & LP_READ)
    {
	if (lp->num_changes == lp->max_changes)
	{
	    lp->max_changes += 10;
	    lp->kq_change =

		REALLOC (lp->kq_change,
			 sizeof (struct kevent) * lp->max_changes);
	}
	memset (&lp->kq_change[lp->num_changes], 0, sizeof (struct kevent));

	lp->kq_change[lp->num_changes].ident = fd;
	lp->kq_change[lp->num_changes].filter = EVFILT_READ;
	lp->kq_change[lp->num_changes].flags = EV_ADD;
	lp->kq_change[lp->num_changes].udata = data;
	lp->num_changes++;
    }
    if (flags & LP_WRITE)
    {
	if (lp->num_changes == lp->max_changes)
	{
	    lp->max_changes += 10;
	    lp->kq_change =

		REALLOC (lp->kq_change,
			 sizeof (struct kevent) * lp->max_changes);
	}
	memset (&lp->kq_change[lp->num_changes], 0, sizeof (struct kevent));

	lp->kq_change[lp->num_changes].ident = fd;
	lp->kq_change[lp->num_changes].filter = EVFILT_WRITE;
	lp->kq_change[lp->num_changes].flags = EV_ADD;
	lp->kq_change[lp->num_changes].udata = data;
	lp->num_changes++;
    }
#elif HAVE_POLL
    /* ensure the offset into the poll_fd[] array looks kosher */
    ASSERT (bucket->pollfd_offset < lp->num_fds);
    pfd = &lp->poll_fd[bucket->pollfd_offset];
    if (flags & LP_READ)
	pfd->events |= POLLIN;
    if (flags & LP_WRITE)
	pfd->events |= POLLOUT;
#else
    if (flags & LP_READ)
	FD_SET (fd, &lp->read_fds);
    if (flags & LP_WRITE)
	FD_SET (fd, &lp->write_fds);
    if (fd > lp->max_fd)
	lp->max_fd = fd;
#endif

    /* set user callback data for this fd */
    bucket->data = data;
}

void
lp_clear (lp_ctx * lp, int fd, int flags)
{
    lp_fd_ctx *bucket;

#if HAVE_POLL && !defined(HAVE_KQUEUE)
    struct pollfd *pfd;
#endif

    ASSERT (lp_validate_ctx (lp));
    bucket = find_fd (lp, fd);
    if (!bucket)
    {
	fprintf (stderr, "lp_clear: descriptor %d was not set\n", fd);
	return;
    }
    ASSERT (lp_validate_fd_ctx (bucket));

#if HAVE_KQUEUE
    if (flags & LP_READ)
    {
	if (lp->num_changes == lp->max_changes)
	{
	    lp->max_changes += 10;
	    lp->kq_change =

		REALLOC (lp->kq_change,
			 sizeof (struct kevent) * lp->max_changes);
	}
	memset (&lp->kq_change[lp->num_changes], 0, sizeof (struct kevent));

	lp->kq_change[lp->num_changes].ident = fd;
	lp->kq_change[lp->num_changes].filter = EVFILT_READ;
	lp->kq_change[lp->num_changes].flags = EV_DELETE;
	lp->num_changes++;
    }
    if (flags & LP_WRITE)
    {
	if (lp->num_changes == lp->max_changes)
	{
	    lp->max_changes += 10;
	    lp->kq_change =

		REALLOC (lp->kq_change,
			 sizeof (struct kevent) * lp->max_changes);
	}
	memset (&lp->kq_change[lp->num_changes], 0, sizeof (struct kevent));

	lp->kq_change[lp->num_changes].ident = fd;
	lp->kq_change[lp->num_changes].filter = EVFILT_WRITE;
	lp->kq_change[lp->num_changes].flags = EV_DELETE;
	lp->num_changes++;
    }
#elif HAVE_POLL
    ASSERT (bucket->pollfd_offset < lp->num_fds);
    ASSERT (fd == lp->poll_fd[bucket->pollfd_offset].fd);
    pfd = &lp->poll_fd[bucket->pollfd_offset];
    if (flags & LP_READ)
    {
	pfd->events &= ~POLLIN;
	pfd->revents &= ~POLLIN;
    }
    if (flags & LP_WRITE)
    {
	pfd->events &= ~POLLOUT;
	pfd->revents &= ~POLLOUT;
    }
#else
    if (flags & LP_READ)
    {
	FD_CLR ((int)fd, &lp->read_fds);
	FD_CLR ((int)fd, &lp->t_read_fds);
    }
    if (flags & LP_WRITE)
    {
	FD_CLR ((int)fd, &lp->write_fds);
	FD_CLR ((int)fd, &lp->t_write_fds);
    }
#endif

    bucket->flags &= ~flags;

    /* if we are no longer waiting on this fd clear it */
    if (bucket->flags == 0)
    {
	lp_fd_ctx **t_bucket;

#if HAVE_POLL && !defined(HAVE_KQUEUE)
	lp_fd_ctx *tmp;

	if (bucket->pollfd_offset < lp->num_fds - 1)
	{
	    struct pollfd *newfd = &lp->poll_fd[lp->num_fds - 1];

	    /* fill the hole created by removing this fd with the one at the
	     * end of the array
	     * must update the .pollfd_offset of the element just moved
	     */
	    tmp = find_fd (lp, newfd->fd);
	    if (tmp)
	    {
		ASSERT (lp_validate_fd_ctx (tmp));
		tmp->pollfd_offset = bucket->pollfd_offset;

		/* copy the struct to its new home */
		memcpy (pfd, newfd, sizeof (struct pollfd));
	    }
	    else
	    {
		fprintf (stderr, "lp_clear: no context for descriptor %d\n",
			 newfd->fd);
	    }
	}
#endif

	/* remove from the hash table.  check by pointer value and not
	 * fd since the fd for poll_fd[] will have been chagned above.
	 */
	t_bucket = &lp->bucket[fd % lp->num_buckets];
	for (; *t_bucket; t_bucket = &(*t_bucket)->next)
	{
	    if (*t_bucket == bucket)
	    {
		*t_bucket = (*t_bucket)->next;
		FREE (bucket);
		break;
	    }
	}

	lp->num_fds--;
    }
}

int
lp_poll (lp_ctx * lp, int timeout /* millisecs */ )
{
#if HAVE_KQUEUE
    struct timespec to;
#elif !defined(HAVE_POLL)
    struct timeval to;
#endif

    ASSERT (lp_validate_ctx (lp));
#if HAVE_KQUEUE
#define ERR "kevent"

    if (timeout >= 0)
    {
	to.tv_sec = timeout / 1000;
	to.tv_nsec = (timeout % 1000) * 100000;	/* nanoseconds */
    }
    /* make sure we have enough room to return all events */
    if (lp->num_fds > lp->max_events)
    {
	lp->max_events = lp->num_fds;
	lp->kq_event =

	    REALLOC (lp->kq_event, sizeof (struct kevent) * lp->max_events);
    }
#if DEBUG
    printf ("lp_poll: %d changes\n", lp->num_changes);
#endif

    lp->num_events = kevent (lp->kq_fd, lp->kq_change, lp->num_changes,
			     lp->kq_event, lp->max_events,
			     (timeout >= 0) ? &to : (struct timespec *)0);
    /* reset */
    lp->num_changes = 0;
#elif HAVE_POLL
#define ERR "poll"
    lp->num_events = poll (lp->poll_fd, lp->num_fds, timeout);
#else
#define ERR "select"

    if (timeout >= 0)
    {
	to.tv_sec = timeout / 1000;
	to.tv_usec = (timeout % 1000) * 1000;
    }
    /* copy the bitmask so we don't destroy the original copy */
    lp->t_read_fds = lp->read_fds;
    lp->t_write_fds = lp->write_fds;

    lp->num_events =
	select (lp->max_fd + 1, &lp->t_read_fds, &lp->t_write_fds, 0,
		(timeout >= 0) ? &to : (struct timeval *)0);
#endif
    if (lp->num_events == -1)
	fprintf (stderr, "lp_poll: " ERR ": %s (errno %d)\n",
		 strerror (errno), errno);
#undef ERR
    return lp->num_events;
}

int
lp_is_set (lp_ctx * lp, int fd, int flags)
{
#if HAVE_KQUEUE
    int     i;
#elif HAVE_POLL
    lp_fd_ctx *bucket;
    struct pollfd *pfd;
#endif

    ASSERT (lp_validate_ctx (lp));

    if (lp->num_events <= 0)
	return 0;		/* nothing pending */
#if HAVE_KQUEUE
    for (i = 0; i < lp->num_events; i++)
    {
	if (lp->kq_event[i].ident == (uintptr_t) fd)
	{
	    if (flags & LP_READ)
	    {
		if (lp->kq_event[i].filter & EVFILT_READ)
		    return 1;
	    }
	    else if (flags & LP_WRITE)
	    {
		if (lp->kq_event[i].filter & EVFILT_WRITE)
		    return 1;
	    }
	    break;
	}
    }
#elif HAVE_POLL
    bucket = find_fd (lp, fd);
    if (!bucket)
    {
	fprintf (stderr, "lp_is_set: no context for descriptor %d", fd);
	return 0;
    }
    ASSERT (lp_validate_fd_ctx (bucket));
    ASSERT (bucket->pollfd_offset < lp->num_fds);
    pfd = &lp->poll_fd[bucket->pollfd_offset];
    if (flags & LP_READ)
	return ((pfd->revents & (POLLIN | POLLRDNORM)) != 0);
    if (flags & LP_WRITE)
	return ((pfd->revents & (POLLOUT | POLLWRNORM)) != 0);
    if (flags & LP_ERROR)
	return ((pfd->revents & (POLLNVAL | POLLERR | POLLHUP)) != 0);
#else
    if (FD_ISSET
	((int)fd, (flags & LP_READ) ? &lp->t_read_fds : &lp->t_write_fds))
	return 1;
#endif
    return 0;
}

void
lp_final (lp_ctx * lp)
{
    ASSERT (lp_validate_ctx (lp));
    lp_zero (lp);
#if HAVE_POLL && !defined(HAVE_KQUEUE)
    FREE (lp->poll_fd);
#endif
    FREE (lp->bucket);
    FREE (lp);
}

/*
 * call the user's callback function for each descriptor which has changed in
 * some manner.
 */
void
lp_dispatch (lp_ctx * lp, lp_callback_t func, void *arg)
{
    int     i;
    int     flags;
    lp_fd_ctx *bucket;

    ASSERT (lp_validate_ctx (lp));
#if DEBUG
    fprintf (stderr, "lp_dispatch: %d pending events\n", lp->num_events);
#endif
    if (lp->num_events <= 0)
	return;			/* no pending events */
#if HAVE_KQUEUE
    (void)bucket;
    for (i = 0; i < lp->num_events; i++)
    {
	/* unfortunately kevent() will return read/write events
	   in separate structures, so the callback could get called
	   twice.  should i attempt to aggregate these or should
	   the callback just assume this can happen?  probably, since
	   iterating over the return list defeats some of the purpose
	   of using kqueue() */
	if (lp->kq_event[i].filter == EVFILT_READ)
	    flags = LP_READ;
	else if (lp->kq_event[i].filter == EVFILT_WRITE)
	    flags = LP_WRITE;
	else
	    flags = 0;
	if (flags)
	    func (lp->kq_event[i].ident, flags, lp->kq_event[i].udata, arg);
    }
    lp->num_events = 0;		/* wait for next lp_poll() call */
#elif HAVE_POLL
    for (i = 0; i < lp->num_fds; i++)
    {
	flags = 0;
	if (lp->poll_fd[i].revents & (POLLIN | POLLRDNORM))
	    flags |= LP_READ;
	if (lp->poll_fd[i].revents & (POLLOUT | POLLWRNORM))
	    flags |= LP_WRITE;
	/* check for errors on the socket.  the callback is responsible for
	 * removing this fd from the set.
	 */
	if (lp->poll_fd[i].revents & (POLLNVAL | POLLERR | POLLHUP))
	{
	    /* invalid descriptor/error/hangup */
	    flags |= LP_ERROR;
	}
	/* check to see if other useful bits might be set.  this is mostly
	 * for debugging so we don't end up consuming 100% cpu
	 */
	if (lp->poll_fd[i].
	    revents & ~(POLLIN | POLLOUT | POLLHUP | POLLERR | POLLNVAL |
			POLLRDNORM | POLLWRNORM))
	{
	    fprintf (stderr,
		     "lp_dispatch: extra bits set on descriptor %d: %u\n",
		     lp->poll_fd[i].fd,
		     lp->poll_fd[i].
		     revents & ~(POLLIN | POLLOUT | POLLHUP | POLLERR |
				 POLLNVAL));
	}
	if (flags)
	{
	    bucket = find_fd (lp, lp->poll_fd[i].fd);
	    if (!bucket)
	    {
		fprintf (stderr,
			 "lp_dispatch: no context for descriptor %d\n",
			 lp->poll_fd[i].fd);
		continue;
	    }
	    ASSERT (lp_validate_fd_ctx (bucket));
	    ASSERT (bucket->pollfd_offset == i);
	    func (lp->poll_fd[i].fd, flags, bucket ? bucket->data : 0, arg);
	}
    }
#else
    for (i = 0; i < lp->num_buckets; i++)
    {
	bucket = lp->bucket[i];
	while (bucket)
	{
	    ASSERT (lp_validate_fd_ctx (bucket));
	    flags = 0;
	    if (FD_ISSET (bucket->fd, &lp->t_read_fds))
		flags |= LP_READ;
	    if (FD_ISSET (bucket->fd, &lp->t_write_fds))
		flags |= LP_WRITE;
	    if (flags)
		func (bucket->fd, flags, bucket->data, arg);
	    bucket = bucket->next;
	}
    }
#endif
}

/* return the current flags for the given fd.  this is just to check what
 * the client is currently watching for.
 */
int
lp_flags (lp_ctx * lp, int fd)
{
    lp_fd_ctx *bucket = find_fd (lp, fd);

#if HAVE_POLL && !defined(HAVE_KQUEUE)
    ASSERT (bucket->pollfd_offset < lp->num_fds);
    ASSERT (fd == lp->poll_fd[bucket->pollfd_offset].fd);
#endif
    return bucket ? bucket->flags : 0;
}
